unit uconsoletimer;

{$mode delphi}{$H+}

interface

uses
  Classes, SyncObjs;

type
  //REMARK DO NOT USE this timer for precision - max res = 1 sec
  TTimerCheck = class(TObject)
  protected
  private
    FFirstTime: Boolean;
    FFireOnFirstTime: Boolean;

    FLastTimer: TDateTime;
    FLastCheck: TDateTime;

    FClockTolerance: Real;
    FAlwaysRaiseOnRevertedClock: Boolean;

    FIngoreClockScrew: Boolean;
    FCheckSteps: Boolean;

  public
    procedure Reset;
    function Check(InterSec: Integer): Boolean;
    constructor Create;
  published
    property ClockTolerance: Real read FClockTolerance write FClockTolerance;
    property AlwaysRaiseOnRevertedClock: Boolean read FAlwaysRaiseOnRevertedClock write FAlwaysRaiseOnRevertedClock;
    property IngoreClockScrew: Boolean read FIngoreClockScrew write FIngoreClockScrew;
    property FireOnFirstTime: Boolean read FFireOnFirstTime write FFireOnFirstTime;
    property CheckSteps: Boolean read FCheckSteps write FCheckSteps;
  end;

  //REMARK DO NOT USE this timer for precision
  TConsoleTimer = class(TThread)
  private
    FExecLock: TCriticalSection;

    FMyTimer: TTimerCheck;
    FSync: Boolean;

    FEnabled: Boolean;
    FInterval: Integer;
    FIntervalSec: Integer;
    FOnTimer: TNotifyEvent;
    FResolution: Integer;
    procedure FireEvent;

    procedure SetEnabled(const Value: Boolean);
    procedure SetOnTimer(const Value: TNotifyEvent);

    procedure SetResolution(const Value: Integer);
    procedure SetInterval(const Value: Integer);

  public
    procedure Execute; override;
    procedure Terminate; reintroduce;

    constructor Create; reintroduce;
    destructor Destroy; override;
  published
    property OnTimer: TNotifyEvent read FOnTimer write SetOnTimer;
    property Interval: Integer read FInterval write SetInterval;
    property Enabled: Boolean read FEnabled write SetEnabled;
    property DoSynchronize: Boolean read FSync write FSync;
    property Resolution: Integer read FResolution write SetResolution;
  end;

implementation

uses SysUtils;

{ TTimerCheck }

function TTimerCheck.Check(InterSec: Integer): Boolean;
var NowTime: TDateTime;
begin
  Result := False;
  if InterSec = 0 then exit;
  NowTime := Now;

  if FFirstTime and FFireOnFirstTime then
  begin
    FFirstTime := False;
    FLastTimer := NowTime;
    FLastCheck := NowTime;
    Result := True;
    exit;
  end;

  if not IngoreClockScrew then
  begin
    //did the clock change in a way we dont want it to change ?
    if FAlwaysRaiseOnRevertedClock and ((NowTime - FLastTimer) < 0) then
    begin
      //we dont allow reverse clock changes per defaut no matter how much
      FLastTimer := NowTime;
      FLastCheck := NowTime;
      Result := True;
      exit;
    end;
    {$ifndef debug}
    if FCheckSteps then
      if abs(NowTime - FLastTimer) > FClockTolerance then
      begin
        //whoowhooo somebody played with the clock
        FLastTimer := NowTime;
        FLastCheck := NowTime;
        Result := True;
        exit;
      end;
    {$endif}
  end;

  FLastTimer := NowTime;
  if ((FLastCheck + (InterSec / SecsPerDay)) < NowTime ) then
  begin
    //here we go
    FLastCheck := NowTime;
    Result := True;
  end;
end;

constructor TTimerCheck.Create;
begin
  FCheckSteps := False;
  FLastTimer := now;
  FLastCheck := FLastTimer;
  FClockTolerance := 10/SecsPerDay; //10 sec tolerance
  AlwaysRaiseOnRevertedClock := False;
  IngoreClockScrew := False;
  FFireOnFirstTime := False;
  FFirstTime := True;
end;

procedure TTimerCheck.Reset;
begin
  FLastCheck := now;
  FLastTimer := FLastCheck;
end;


{ TConsoleTimer }

constructor TConsoleTimer.Create;
begin
  FSync := True;
  FOnTimer := nil;
  FMyTimer := TTimerCheck.Create;
  FMyTimer.CheckSteps := True;
  FExecLock := TCriticalSection.Create;
  FResolution := 100;
  inherited Create(false);
  FreeOnTerminate := True;
end;

destructor TConsoleTimer.Destroy;
begin
  FOnTimer := nil;
  FExecLock.Enter;
  FExecLock.Leave;

  FMyTimer.Free;
  FExecLock.Free;

  inherited;
end;

procedure TConsoleTimer.execute;
begin
  while not Terminated do
  begin
    //one sec we can handle with checking the clock
    if (FInterval >= 1000) then
    begin
      if FEnabled and assigned(FOnTimer) then
      begin
        FExecLock.Enter;
        try
          //WriteLn('FMyTimer.Check ', FIntervalSec);
          if FMyTimer.Check(FIntervalSec) then
            if FSync then
              Synchronize(FireEvent)
            else
              FireEvent;
        finally
          FExecLock.Leave;
        end;
      end;
      Sleep(FResolution);
    end else
    begin
      //else we simply sleep
      //very buggy compared to checking the time but it should not matter on this small timings
      //REMARK DO NOT USE this timer for precision
      if FEnabled then
      begin
        Sleep(FInterval);
        if FSync then
          Synchronize(FireEvent)
        else
          FireEvent;
      end else
        Sleep(FResolution);
    end;
  end;
end;

procedure TConsoleTimer.FireEvent;
begin
  if Assigned(FOnTimer) then
    FOnTimer(Self);
end;

procedure TConsoleTimer.SetEnabled(const Value: Boolean);
begin
  if FEnabled = Value then Exit;
  FEnabled := Value;
  FMyTimer.Reset;
end;

procedure TConsoleTimer.SetOnTimer(const Value: TNotifyEvent);
begin
  FExecLock.Enter;
  try
    FOnTimer := Value;
  finally
    FExecLock.Leave;
  end;
end;

procedure TConsoleTimer.SetResolution(const Value: Integer);
begin
  FResolution := Value;
  if FResolution > FInterval then
    FResolution := FInterval;
end;

procedure TConsoleTimer.SetInterval(const Value: Integer);
begin
  FInterval := Value;
  FIntervalSec := Value div 1000;
end;

procedure TConsoleTimer.Terminate;
begin
  inherited;
  FOnTimer := nil;
end;


end.

