unit uwinevent;

interface

uses
  Classes;
  

const
  INFINITE = $FFFFFF;

  STATUS_WAIT_0  = $00000000;
  STATUS_TIMEOUT = $00000102;

  WAIT_TIMEOUT = STATUS_TIMEOUT;
  WAIT_FAILED = DWORD($FFFFFFFF);
  WAIT_OBJECT_0 = ((STATUS_WAIT_0) + 0);

  MAXIMUM_WAIT_OBJECTS = 64;
  
  INVALID_HANDLE_VALUE = 0;
  
type
  PSecurityAttributes = Pointer;

  TWOHandleArray = Array[0 .. MAXIMUM_WAIT_OBJECTS - 1] of THandle;
  PWOHandleArray = ^TWOHandleArray;
  
  TLargeInteger = Int64;


  function WaitForSingleObject(aEvent : THandle; aTimeOut : DWord) :  DWORD;
  function WaitForMultipleObjects(nCount : DWord; lpHandles : PWOHandleArray; bWaitAll : Boolean; dwMilliseconds : DWord) : DWORD;

  function CloseHandle(hObject: THandle) : Boolean;
  
  (* Semaphoren *)
  function CreateSemaphore(lpSemaphoreAttributes: PSecurityAttributes; lInitialCount, lMaximumCount : LongInt; lpName : PChar) : THandle;
  function ReleaseSemaphore(hSemaphore : THandle; lReleaseCount : LongInt; lpPreviousCount: Pointer) : Boolean;
  function _debugSemaphoreSignalCnt(hSemaphore : THandle) : Integer;
  
  (* Event *)
  function CreateEvent(lpEventAttributes: PSecurityAttributes; bManualReset, bInitialState : Boolean; lpName : PChar) : THandle;
  function SetEvent(hEvent : THandle): Boolean;
  function ResetEvent(hEvent : THandle): Boolean;

  (* Waitable Timer Functions *)
  function CreateWaitableTimer(lpTimerAttributes : PSecurityAttributes; bManualReset : Boolean; lpTimerName : PChar) : THandle;
  function KillTimer(hWnd, uIDEvent : THandle) : Boolean;
  function SetWaitableTimer(hTimer : THandle; var lpDueTime: TLargeInteger; lPeriod : LongInt;
    pfnCompletionRoutine, lpArgToCompletionRoutine : Pointer; fResume : Boolean) : Boolean;


implementation

uses
  SysUtils, uWinEventHandle;
  
resourcestring
  SSemaphoreIsNil = 'Semaphore can''t released its nil';
  SPreviousCountNotSupported = 'PreviousCount at ReleaseSemaphore not supported';
  SEventIsNil = 'Event can''t set its nil';
  SRestEventIsNil = 'Event can''t rest its nil';
  SNoManualResetForTimer = 'Manualreset for timers not supported';

function WaitForSingleObject(aEvent: THandle; aTimeOut: DWord): DWORD;
Var
  aBaseEvent : TBaseWinEventHandle;
begin
  aBaseEvent := TBaseWinEventHandle(aEvent);
  if aBaseEvent.Wait(aTimeOut) then
    Result := WAIT_OBJECT_0
  else
    Result := WAIT_TIMEOUT;
end;

function WaitForMultipleObjects(nCount: DWord; lpHandles: PWOHandleArray;
  bWaitAll: Boolean; dwMilliseconds: DWord): DWORD;
var
  I : Integer;
  aSubcEvent : TSubscriberEvent;
  aSignaledEvent : THandle;
begin
  Result := WAIT_FAILED;
  Assert(not bWaitAll, 'Not impleacmented');
    
  //Check whether an event is already signaled
  for I := 0 to nCount -1 do
  begin
    Assert(lpHandles^[I] <> 0, 'Event is nil');
    if TBaseWinEventHandle(lpHandles^[I]).Wait(DWord(-1)) then
    begin
      Result := WAIT_OBJECT_0 + I;
      exit;
    end;
  end;
  
  //Now the trick: no event signaled yet, create a subscriber event;
  aSubcEvent := TSubscriberEvent.Create;
  try
    for I := 0 to nCount -1 do
      TBaseWinEventHandle(lpHandles^[I]).AddSubscribers(aSubcEvent);
      
    aSignaledEvent := aSubcEvent.Wait(dwMilliseconds);
    
    for I := 0 to nCount -1 do
      TBaseWinEventHandle(lpHandles^[I]).RemoveSubscribers(aSubcEvent);

    if aSignaledEvent <> 0 then
    begin
      for I := 0 to nCount -1 do
        if lpHandles^[I] = aSignaledEvent then
        begin
          Result := WAIT_OBJECT_0 + I;
          break;
        end;
    end else
      Result := WAIT_TIMEOUT;
  finally
    aSubcEvent.free;
  end;
end;

function CloseHandle(hObject: THandle): Boolean;
begin
  Result := true;
  TObject(hObject).Free;
end;

function CreateSemaphore(lpSemaphoreAttributes: PSecurityAttributes;
  lInitialCount, lMaximumCount: LongInt; lpName: PChar): THandle;
begin
  Result := THandle( TWinSemaphore.Create(lInitialCount, lMaximumCount) );
end;

function ReleaseSemaphore(hSemaphore: THandle; lReleaseCount: LongInt;
  lpPreviousCount: Pointer): Boolean;
begin
  Result := true;
  if hSemaphore = 0 then
    raise Exception.Create(SSemaphoreIsNil);
  if lpPreviousCount <> nil then
    raise Exception.Create(SPreviousCountNotSupported);
  TWinSemaphore(hSemaphore).Signal(lReleaseCount);
end;

function _debugSemaphoreSignalCnt(hSemaphore: THandle): Integer;
begin
  Result := TWinSemaphore(hSemaphore).SigCnt;
end;

function CreateEvent(lpEventAttributes: PSecurityAttributes; bManualReset,
  bInitialState: Boolean; lpName: PChar): THandle;
begin
  Result := THandle( TWinEvent.Create(bInitialState, bManualReset) );
end;

function SetEvent(hEvent: THandle): Boolean;
begin
  if hEvent = 0 then
    raise Exception.Create(SEventIsNil);
  Result := false;
  if not TWinEvent(hEvent).Signaled then
    TWinEvent(hEvent).Signal;
end;

function ResetEvent(hEvent: THandle): Boolean;
begin
  if hEvent = 0 then
    raise Exception.Create(SRestEventIsNil);
  TWinEvent(hEvent).Reset;
  Result := true;
end;

function CreateWaitableTimer(lpTimerAttributes: PSecurityAttributes;
  bManualReset: Boolean; lpTimerName: PChar): THandle;
begin
  if bManualReset then
    raise Exception.Create(SNoManualResetForTimer);
  Result := THandle( TWinWaitableTimer.Create(bManualReset) );
end;

function KillTimer(hWnd, uIDEvent: THandle): Boolean;
begin
  Result := true;
  TWinWaitableTimer( uIDEvent ).KillTimer;
end;

function SetWaitableTimer(hTimer: THandle; var lpDueTime: TLargeInteger;
  lPeriod: LongInt; pfnCompletionRoutine, lpArgToCompletionRoutine: Pointer;
  fResume: Boolean): Boolean;
begin
  Result := true;
  TWinWaitableTimer( hTimer ).SetTimer(lPeriod);
end;


end.
