unit uWinEventHandle;

interface

{$MODE DELPHI}

uses
  Classes, uconsoletimer;
  
type
  TSubscriberEvent = class;

  { TBaseWinEventHandle }

  TBaseWinEventHandle = Class(TObject)
  private
    fSubscribers : TThreadList;
    fEvent : PRTLEvent;
    _Count : Integer;
    
  protected
    function SignalSubscribers : Boolean; virtual;
    function RtlWait(aNumMs : DWord) : Boolean; virtual;
    
  public
    constructor Create(aInitCnt : Integer);
    destructor Destroy; override;
    
    procedure Reset; virtual; abstract;
    function Wait(aNumMs : DWord) : Boolean; virtual; abstract;

    function Signaled : Boolean; virtual; abstract;
    procedure Signal(aCount : Integer = 1); virtual; abstract;
    
    procedure AddSubscribers(aSubc : TSubscriberEvent); virtual;
    procedure RemoveSubscribers(aSubc : TSubscriberEvent); virtual;
  end;
  
  { TSubscriberEvent }

  TSubscriberEvent = Class(TBaseWinEventHandle)
  public
    constructor Create;
    
    procedure Reset; override;
    function Wait(aNumMs : DWord) : THandle; reintroduce;

    function Signaled : Boolean; override;
    procedure Signal(aHandle : Integer); override;

    procedure AddSubscribers(aSubc : TSubscriberEvent); override;
    procedure RemoveSubscribers(aSubc : TSubscriberEvent); override;
  end;

  { TWinSemaphore }

  TWinSemaphore = Class(TBaseWinEventHandle)
  private
    fMaxCnt : Integer;
  public
    constructor Create(aInitCnt, aMaxCnt : Integer);

    procedure Reset; override;
    function Wait(aNumMs : DWord) : Boolean; override;

    function SigCnt : Integer;
    function Signaled : Boolean; override;
    procedure Signal(aCount : Integer = 1); override;
  end;
  
  { TWinEvent }

  TWinEvent = Class(TBaseWinEventHandle)
  private
    fManualReset : Boolean;
  public
    constructor Create(aInitState, aManualReset : Boolean);

    procedure Reset; override;
    function Wait(aNumMs : DWord) : Boolean; override;

    function Signaled : Boolean; override;
    procedure Signal; reintroduce;
  end;
  
  
  { TWinWaitableTimer }

  TWinWaitableTimer = Class(TWinSemaphore)
  private
    fTimer : TConsoleTimer;
    fManualReset : Boolean;
  protected
    procedure TimerSignal(Sender: TObject);

  public
    constructor Create(aManualReset : Boolean);
    destructor Destroy; override;
    
    procedure KillTimer;
    procedure SetTimer(aInterval: DWord);
  end;

implementation

uses
  SysUtils, uwinevent;


//--------------------------------------------------------------------------------------------------
// Locked Integer manipulation
//
// Routines to manipulate simple typed variables in a thread safe manner
//--------------------------------------------------------------------------------------------------

function LockedAdd(var Target: Integer; Value: Integer): Integer; assembler;
asm
        MOV     ECX, EAX
        MOV     EAX, EDX
        LOCK XADD [ECX], EAX
        ADD     EAX, EDX
end;

function LockedDec(var Target: Integer): Integer; assembler;
asm
        MOV     ECX, EAX
        MOV     EAX, -1
        LOCK XADD [ECX], EAX
        DEC     EAX
end;

function LockedExchange(var Target: Integer; Value: Integer): Integer; assembler;
asm
        MOV     ECX, EAX
        MOV     EAX, EDX
        LOCK XCHG [ECX], EAX
end;

function LockedExchangeAdd(var Target: Integer; Value: Integer): Integer; assembler;
asm
        MOV     ECX, EAX
        MOV     EAX, EDX
        LOCK XADD [ECX], EAX
end;


{ TBaseWinEventHandle }

constructor TBaseWinEventHandle.Create(aInitCnt: Integer);
begin
  inherited Create;
  _Count := aInitCnt;
  fSubscribers := TThreadList.Create;
  fEvent := RtlEventCreate;
  RtlEventWaitFor(fEvent, 1); //TODO Patch 2.1.1 RtlEventSetEvent Raise AV if call RTLSetEvent before called RTLEventStartWait
end;

destructor TBaseWinEventHandle.Destroy;
begin
  RtlEventDestroy(fEvent);
  fSubscribers.Free;
  inherited Destroy;
end;

procedure TBaseWinEventHandle.AddSubscribers(aSubc: TSubscriberEvent);
begin
  fSubscribers.Add(aSubc);
end;

procedure TBaseWinEventHandle.RemoveSubscribers(aSubc: TSubscriberEvent);
begin
  fSubscribers.Remove(aSubc);
end;

function TBaseWinEventHandle.SignalSubscribers : Boolean;
Var
  I : Integer;
  aList : TList;
begin
  aList := fSubscribers.LockList;
  try
    Result := aList.Count <> 0;
    for I := 0 to aList.Count -1 do
      TSubscriberEvent(aList[I]).Signal(THandle(Self));
  finally
    fSubscribers.UnlockList;
  end;
end;

function TBaseWinEventHandle.RtlWait(aNumMs: DWord): Boolean;
begin
  if not Signaled and (aNumMs <> DWord(-1)) then  //aNumMs mit -1 wird von WaitForMultiple um
  begin                                           //zu schauen ob ein event schon ausgelst wurde
    if aNumMs <> INFINITE then
      RTLeventWaitFor(fEvent, aNumMs)
    else
      RTLeventWaitFor(fEvent);
  end;
  Result := Signaled;
end;

{ TSubscriberEvent }

constructor TSubscriberEvent.Create;
begin
  inherited Create(0);
end;

function TSubscriberEvent.Signaled: Boolean;
begin
  Result:= LockedExchangeAdd(_Count, 0) <> 0;
end;

procedure TSubscriberEvent.Reset;
begin
  Assert(false, 'SubscriberEvent cant reset');
end;

function TSubscriberEvent.Wait(aNumMs: DWord): THandle;
begin
  if not Signaled then
  begin
    if aNumMs <> INFINITE then
      RTLeventWaitFor(fEvent, aNumMs)
    else
      RTLeventWaitFor(fEvent);
  end;
  Result := LockedExchangeAdd(_Count, 0);
end;

procedure TSubscriberEvent.Signal(aHandle : Integer);
begin
  if Signaled then
    Exit;
  LockedExchange(_Count, aHandle);
  TBaseWinEventHandle(aHandle).Wait(DWord(-1));
  RtlEventSetEvent(fEvent);
end;

procedure TSubscriberEvent.AddSubscribers(aSubc: TSubscriberEvent);
begin
  Assert(false, 'An SubscriberEvent cant subcribe an other');
end;

procedure TSubscriberEvent.RemoveSubscribers(aSubc: TSubscriberEvent);
begin
  Assert(false, 'An SubscriberEvent cant subcribe an other');
end;

{ TWinSemaphore }

constructor TWinSemaphore.Create(aInitCnt, aMaxCnt: Integer);
begin
  inherited Create(aInitCnt);
  fMaxCnt := aMaxCnt;
end;

procedure TWinSemaphore.Reset;
begin
  LockedDec(_Count);
end;

function TWinSemaphore.Wait(aNumMs: DWord): Boolean;
begin
  Result := RtlWait(aNumMs);
  if Result then
    Reset;
end;

function TWinSemaphore.SigCnt: Integer;
begin
  Result:= LockedExchangeAdd(_Count, 0);
end;

function TWinSemaphore.Signaled: Boolean;
begin
  Result:= LockedExchangeAdd(_Count, 0) <> 0;
end;

procedure TWinSemaphore.Signal(aCount: Integer);
begin
  LockedAdd(_Count, aCount);
  RtlEventSetEvent(fEvent);
  SignalSubscribers;
end;

{ TWinEvent }

constructor TWinEvent.Create(aInitState, aManualReset: Boolean);
Var
  aInit : Integer;
begin
  if aInitState then
    aInit := 1
  else
    aInit := 0;
  inherited Create(aInit);
  fManualReset := aManualReset;
end;

function TWinEvent.Signaled: Boolean;
begin
  Result:= LockedExchangeAdd(_Count, 0) = 1;
end;

procedure TWinEvent.Reset;
begin
  LockedExchange(_Count, 0);
end;

function TWinEvent.Wait(aNumMs: DWord): Boolean;
begin
  Result := RtlWait(aNumMs);
  if Result and  not fManualReset then
    Reset;
end;

procedure TWinEvent.Signal;
begin
  LockedExchange(_Count, 1);
  RtlEventSetEvent(fEvent);
  SignalSubscribers;
end;


{ TWinWaitableTimer }

constructor TWinWaitableTimer.Create(aManualReset: Boolean);
begin
  inherited Create(0, $FFFF);
  fTimer:= TConsoleTimer.Create;
  fTimer.Resolution := 1000;
  fTimer.DoSynchronize := False;
  fTimer.Enabled := false;
  fManualReset := aManualReset;
end;

destructor TWinWaitableTimer.Destroy;
begin
  KillTimer;
  fTimer.Terminate;
  inherited Destroy;
end;

procedure TWinWaitableTimer.SetTimer(aInterval: DWord);
begin
  fTimer.OnTimer := TimerSignal;
  fTimer.Interval := aInterval;
  fTimer.Enabled := True;
end;

procedure TWinWaitableTimer.KillTimer;
begin
  fTimer.Enabled := False;
  fTimer.OnTimer := nil;
end;

procedure TWinWaitableTimer.TimerSignal(Sender: TObject);
begin
  Signal;
end;


end.
